/*
 * $Id: DelegationDNSCache.java 249 2005-08-26 18:44:12Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.cache;

import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.xbill.DNS.Credibility;
import org.xbill.DNS.DClass;
import org.xbill.DNS.Flags;
import org.xbill.DNS.NSRecord;
import org.xbill.DNS.Name;
import org.xbill.DNS.Section;
import org.xbill.DNS.Type;

import se.rfc.unbound.Request;
import se.rfc.unbound.SMessage;
import se.rfc.unbound.SRRset;
import se.rfc.unbound.Util;

/**
 * This is a "delegation-only" cache. That is, it only caches delegation point
 * information, and no actual answers. As such, the lookup() method will never
 * return anything, but the findDelegation() will return the closest cached
 * delegation point.
 * 
 * @author davidb
 * @version $Revision: 249 $
 */
public class DelegationDNSCache implements DNSCache
{

  /**
   * This is a basic cache entry data structure.
   * 
   * @author davidb
   * @version $Revision: 249 $
   */
  private static class Entry
  {
    // the absolute time after which this entry should be discarded.
    public Date   expiration;
    // the credibility of this entry.
    public int    credibility;
    // The rrset being cached!
    public SRRset rrset;

    public Entry(SRRset rrset, long maxTTL, int cred)
    {
      this.rrset = rrset;
      this.credibility = cred;
      long ttl = rrset.getTTL();
      if (maxTTL > 0 && ttl > maxTTL)
      {
        ttl = maxTTL;
        rrset.setTTL(ttl);
      }

      this.expiration = new Date(System.currentTimeMillis() + (ttl * 1000));
    }

    public String toString()
    {
      return rrset.getName() + "/" + DClass.string(rrset.getDClass()) + "/"
          + Type.string(rrset.getType());
    }
  }

  // This is the main caching data structure.
  private CacheCore mCache;
  // The maximum TTL to honor -- TTLs greater than this will be trimmed.
  private int       mMaxTTL = 1209600;

  private Logger    log     = Logger.getLogger(this.getClass());

  public DelegationDNSCache()
  {
    mCache = new CacheCore();
  }

  /**
   * Initialize the cache.
   * 
   * This cache recognizes the following configuration parameters:
   * <dl>
   * <dt>dns.root_hints_file</dt>
   * <dd>The file name for the root hints file</dd>
   * <dt>dns.cache.max_ttl</dt>
   * <dd>The maximum TTL (in seconds) to honor</dd>
   * <dt>dns.cache.max_size</dt>
   * <dd>The maximum number of object that the cache may store.</dd>
   * </dl>
   */
  public void init(Properties config) throws Exception
  {
    if (config == null) return;
    
    String s;

    s = config.getProperty("dns.cache.max_ttl");
    mMaxTTL = Util.parseInt(s, mMaxTTL);

    s = config.getProperty("dns.cache.max_size");
    
    if (s != null)
    {
      int max_size = Integer.parseInt(s);
      mCache.setMaxCacheSize(max_size);
    }
  }

  /*
   * Since this is a "delegation-only" cache, this always returns null.
   */
  public SMessage lookup(Name qname, int qtype, int qclass)
  {
    return null;
  }

  /*
   * Store a given message in the cache. This cache will effectively only
   * store A and NS rrsets.
   */
  public void store(SMessage response, Request originalRequest, boolean isReferral)
  {
    boolean auth_answer = response.getHeader().getFlag(Flags.AA);
    for (int section = Section.ANSWER; section <= Section.ADDITIONAL; section++)
    {
      SRRset[] rrsets = response.getSectionRRsets(section);

      for (int i = 0; i < rrsets.length; i++)
      {
        int type = rrsets[i].getType();
        if (type == Type.NS || type == Type.A || type == Type.AAAA)
        {
          storeRRset(rrsets[i], sectionToCredibility(section, auth_answer));
        }
      }
    }
  }

  /**
   * Given a section number, return a rough estimate of the credibility of the
   * RRset in question.
   * 
   * @param section the section in question (ANSWER -> ADDITIONAL)
   * 
   * @return the credibility number.
   */
  private int sectionToCredibility(int section, boolean auth_answer)
  {
    if (section == Section.ADDITIONAL) return Credibility.ADDITIONAL;
    if (section == Section.AUTHORITY)
    {
      if (auth_answer) return Credibility.AUTH_AUTHORITY;
      return Credibility.NONAUTH_AUTHORITY;
    }
    if (section == Section.ANSWER)
    {
      if (auth_answer) return Credibility.AUTH_ANSWER;
      return Credibility.NONAUTH_ANSWER;
    }
    // Otherwise
    return Credibility.HINT;
  }

  /**
   * Find the closest delegation poit to qname in the cache.
   * 
   * @param qname the query name.
   * @param qclass the query class (usually DClass.IN)
   * 
   * @return a DelegationPoint for the closest delegation, or null if for some
   *         bizarre reason the cache doesn't even have the root delegation
   *         point.
   */
  public SMessage findDelegation(Name qname, int qclass)
  {
    Name n = qname;

    Entry ns_entry = null;
    while (n.labels() > 0)
    {
      ns_entry = lookupRRset(n, Type.NS, qclass);
      if (ns_entry != null) break;
      n = new Name(n, 1);
    }

    // NOTE: this should never actually execute, since we should always find
    // (at least) an NS for root.
    if (ns_entry == null)
    {
      log.error("Returning null because there is no "
          + "root delegation cached (apparently)");
      return null;
    }
    SRRset ns_rrset = ns_entry.rrset;
    List rrset_list = new LinkedList();
    rrset_list.add(ns_rrset);

    for (Iterator i = ns_rrset.rrs(); i.hasNext();)
    {
      NSRecord ns = (NSRecord) i.next();
      Entry entry = lookupRRset(ns.getTarget(), Type.A, qclass);
      if (entry != null)
      {
        rrset_list.add(entry.rrset);
      }
      entry = lookupRRset(ns.getTarget(), Type.AAAA, qclass);
      if (entry != null)
      {
        rrset_list.add(entry.rrset);
      }
    }

    SMessage m = CacheUtils.referralFromList(rrset_list);
    m.getHeader().unsetFlag(Flags.AA);
    m.getHeader().setFlag(Flags.QR);
    
    return m;
  }

  private void storeRRset(SRRset rrset, int cred)
  {
    Entry entry = new Entry(rrset, mMaxTTL, cred);
    String k = key(rrset);

    Entry oldEntry = lookupRRset(k);
    // Only store the RRset in the cache if it wasn't in there already (or
    // expired) or this is more credible than the existing entry.
    if (oldEntry == null || oldEntry.credibility <= cred)
    {
      // log.debug("storing RRset: " + entry.rrset + "at key: " + k);
      CacheUtils.trimRRsetTTL(rrset, null);
      mCache.put(k, entry);
    }
    else
    {
      log.debug("NOT storing RRset: " + entry.rrset.getName() + "/"
          + Type.string(entry.rrset.getType())
          + " because credibility of old entry was " + oldEntry.credibility
          + " compared to " + entry.credibility);
    }
  }

  private Entry lookupRRset(String k)
  {
    Entry entry = (Entry) mCache.get(k);
    if (entry == null) return null;
    // throw out the response if it has expired.
    if (entry.expiration.before(new Date()))
    {
      log.debug("discarding expired cache entry: " + entry + " (expired "
          + entry.expiration + ")");
      mCache.remove(k);
      return null;
    }
    
    // Clone the cached RRset, lowering the TTL.
    SRRset orig_rrset = entry.rrset;
    long new_ttl = CacheUtils.cacheTTL(entry.expiration, orig_rrset.getTTL(), null);
    entry.rrset = orig_rrset.cloneSRRset(new_ttl);
    
    return entry;
  }

  private Entry lookupRRset(Name qname, int qtype, int qclass)
  {
    String k = key(qname, qtype, qclass);
    return lookupRRset(k);
  }

  private String nameToString(Name name)
  {
    String n = name.toString().toLowerCase();
    if (n.endsWith(".")) n = n.substring(0, n.length() - 1);

    return n;
  }

  private String key(SRRset rrset)
  {
    return "R" + rrset.getType() + nameToString(rrset.getName());
  }

  private String key(Name qname, int qtype, int qclass)
  {
    return "R" + qtype + nameToString(qname);
  }

}