/*
 * $Id: HybridDNSCache.java 249 2005-08-26 18:44:12Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.cache;

import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;
import org.xbill.DNS.Credibility;
import org.xbill.DNS.Header;
import org.xbill.DNS.Name;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.Record;
import org.xbill.DNS.Section;
import org.xbill.DNS.Type;

import se.rfc.unbound.Request;
import se.rfc.unbound.SMessage;
import se.rfc.unbound.SRRset;
import se.rfc.unbound.SecurityStatus;
import se.rfc.unbound.Util;

/**
 * This is a "full" DNS cache -- it will do both positive and negative caching
 * of DNS messages. This cache takes a hybrid (between 'per-message' and
 * 'per-rrset') caching approach, based on an idea by Matt Larson.
 * 
 * Basically each message stored will be stored more or less "as-is", except
 * that the RRsets that form the message will be only stored once in the
 * cache, even though they may appear in multiple messages.
 * 
 * @author davidb
 * @version $Revision: 249 $
 */
public class HybridDNSCache implements DNSCache
{
  /**
   * A cached (S)RRset entry.
   * 
   * @author davidb
   * @version $Revision: 249 $
   */
  private static class RRsetEntry
  {
    public Date   expiration;
    public int    credibility;
    public SRRset rrset;

    public RRsetEntry(SRRset rrset, long maxTTL, int cred)
    {
      this.rrset = rrset;
      this.credibility = cred;
      long ttl = rrset.getTTL();
      if (maxTTL > 0 && ttl > maxTTL)
      {
        ttl = maxTTL;
        rrset.setTTL(ttl);
      }

      this.expiration = new Date(System.currentTimeMillis() + (ttl * 1000));
    }

    public String toString()
    {
      return "RRset " + rrset.getName() + "/" + Type.string(rrset.getType())
          + "/" + DClass.string(rrset.getDClass())
          + (rrset.isSigned() ? " (signed)" : "") + " cred[" + credibility
          + "]";
    }
  }

  /**
   * A cached message.
   * 
   * @author davidb
   * @version $Revision: 249 $
   */
  private static class MessageEntry
  {
    public Date           expiration;
    public Header         header;
    public Record         question;
    public SecurityStatus securityStatus;

    // The message sections are really list of cache keys, not the cached
    // objects themselves. This allows the particular RRsets to be updated
    // with later copies, if necessary.
    public String[][]     section;

    public MessageEntry()
    {
      section = new String[3][];
    }

    public String toString()
    {
      if (question == null) return super.toString();
      return "Message " + question.getName() + "/"
          + Type.string(question.getType()) + "/"
          + DClass.string(question.getDClass());
    }
  }

  /** the main cache data structure */
  private CacheCore   mCache;

  /** the maximum TTL to allow. */
  private long        mMaxTTL                = 1209600;

  /** the TTL to use for negative caching when no SOA record is present. */
  private long        mDefaultNegTTL         = 300;

  private Logger      log                    = Logger.getLogger(this
                                                 .getClass());

  private static long DEFAULT_MAX_CACHE_SIZE = 2147483;

  public HybridDNSCache()
  {
    mCache = new CacheCore();
  }

  public void init(Properties config) throws Exception
  {
    if (config == null) return;

    // We have two major tasks here: 1) get the SBELT configured (this is the
    // root hints, normally), 2) configure any other options.

    String s;

    s = config.getProperty("dns.cache.max_ttl");
    mMaxTTL = Util.parseLong(s, mMaxTTL);

    s = config.getProperty("dns.cache.max_size");
    mCache.setMaxCacheSize(Util.parseLong(s, DEFAULT_MAX_CACHE_SIZE));

    s = config.getProperty("dns.cache.default_negative_ttl");
    mDefaultNegTTL = Util.parseLong(s, mDefaultNegTTL);
  }

  public SMessage lookup(Name qname, int qtype, int qclass)
  {
    MessageEntry me = null;
    SMessage m = null;

    // For now, we first lookup the NXDOMAIN cache entry -- this way, a cached
    // NXDOMAIN will supercede any cached message that might match.
    String k = nxMessageKey(qname, qclass);
    me = lookupMessage(k);

    // If we have an NXDOMAIN entry, then make sure that that can be made into
    // an entire message, and return it.
    if (me != null)
    {
      m = messageFromEntry(k, me);
      if (m != null)
      {
        return m;
      }
    }

    // Next, we lookup the CNAME type for this key
    k = messageKey(qname, Type.CNAME, qclass);
    me = lookupMessage(k);
    if (me != null)
    {
      return messageFromEntry(k, me);
    }

    // Otherwise, attempt to look up the actual message.
    k = messageKey(qname, qtype, qclass);
    me = lookupMessage(k);

    if (me != null)
    {
      m = messageFromEntry(k, me);
      if (m != null)
      {
        return m;
      }
    }

    // We special-case DS and DNSKEY requests.
    // If we don't have explictly cached messages for those queries, we
    // attempt to construct responses.
    if (qtype == Type.DS)
    {
      m = lookupDSMessage(qname, qclass);
    }
    else if (qtype == Type.DNSKEY)
    {
      m = lookupDNSKEYMessage(qname, qclass);
    }

    return m;
  }

  public void store(SMessage response, Request request, boolean isReferral)
  {
    MessageEntry me = new MessageEntry();
    me.header = (Header) response.getHeader().clone();
    me.question = response.getQuestion();

    me.securityStatus = response.getSecurityStatus();

    boolean auth = me.header.getFlag(Flags.AA);
    me.header.unsetFlag(Flags.AA);
    long minTTL = -1;

    // Store (or attempt to store) all of the RRsets that make up this
    // message.
    for (int s = Section.ANSWER; s <= Section.ADDITIONAL; s++)
    {
      SRRset[] rrsets = response.getSectionRRsets(s);
      String[] keys = new String[rrsets.length];
      for (int i = 0; i < rrsets.length; i++)
      {
        if (minTTL < 0 || rrsets[i].getTTL() < minTTL)
          minTTL = rrsets[i].getTTL();
        keys[i] = storeRRset(rrsets[i], credibility(s, auth));
      }
      me.section[s - 1] = keys;
    }
    
    // Only ANSWER responses are stored as messages -- REFERRALs are
    // always constructed from RRsets.
    if (isReferral) return;

    // if the message contains a TTL==0 rrset then we do not cache the message
    // entry (possibly orphaning the SRRsets that we *did* cache -- if they
    // are of type NS or A/AAAA then they *might* be used by
    // findDelegation().)
    if (minTTL <= 0) return;

    String k;
    if (response.getRcode() == Rcode.NXDOMAIN)
    {
      me.question = response.getQuestion();
      me.expiration = new Date(System.currentTimeMillis()
          + (negativeCacheTTL(response, minTTL) * 1000));
      k = nxMessageKey(me.question.getName(), me.question.getDClass());
    }
    else
    {
      // Cache answers with their original query
      // The response should already have had the CNAME/DNAME chain prepended!
      me.question = request.getQuestion();
      k = messageKey(me.question.getName(),
          me.question.getType(),
          me.question.getDClass());

      // distinguish here between a positive and negative response (neg.
      // responses are NOERROR/NODATA) responses. we can do this just be
      // noticing if there is anything in the ANSWER
      // section (if response type is ANSWER).
      if (me.header.getCount(Section.ANSWER) == 0)
      {
        me.expiration = new Date(System.currentTimeMillis()
            + (negativeCacheTTL(response, minTTL) * 1000));
      }
      else
      {
        me.expiration = new Date(System.currentTimeMillis() + (minTTL * 1000));
      }

    }
    // And put the message entry into the cache.
    mCache.put(k, me);
  }

  public SMessage findDelegation(Name qname, int qclass)
  {
    Name n = qname;

    RRsetEntry ns_entry = null;
    while (n.labels() > 0)
    {
      ns_entry = lookupRRset(rrsetKey(n, Type.NS, qclass));
      if (ns_entry != null) break;
      n = new Name(n, 1);
    }

    if (ns_entry == null)
    {
      // Hopefully, the resolver has a recourse! (I.e., the resolver's hints)
      return null;
    }

    SRRset ns_rrset = ns_entry.rrset;

    return constructReferral(ns_rrset, qclass);
  }

  private SMessage constructReferral(SRRset ns_rrset, int qclass)
  {
    Name n = ns_rrset.getName();
    List rrset_list = new LinkedList();
    rrset_list.add(ns_rrset);

    // Lookup the DS or NSEC at the delegation point.
    RRsetEntry dnssec_entry = lookupRRset(rrsetKey(n, Type.DS, qclass));
    if (dnssec_entry == null)
    {
      // NOTE: this won't work for alternate NSEC schemes (opt-in, NSEC3).
      dnssec_entry = lookupRRset(rrsetKey(n, Type.NSEC, qclass));
    }
    if (dnssec_entry != null)
    {
      rrset_list.add(dnssec_entry.rrset);
    }

    for (Iterator i = ns_rrset.rrs(); i.hasNext();)
    {
      NSRecord ns = (NSRecord) i.next();
      RRsetEntry entry = lookupRRset(rrsetKey(ns.getTarget(), Type.A, qclass));
      if (entry != null)
      {
        rrset_list.add(entry.rrset);
      }
      entry = lookupRRset(rrsetKey(ns.getTarget(), Type.AAAA, qclass));
      if (entry != null)
      {
        rrset_list.add(entry.rrset);
      }
    }

    SMessage m = CacheUtils.referralFromList(rrset_list);
    m.getHeader().unsetFlag(Flags.AA);
    m.getHeader().setFlag(Flags.QR);

    return m;
  }

  /**
   * Given qname/qclass, attempt to construct a DS response from the cache.
   * This will possibly be missing auth and additional section records, if
   * they expired/were never cached.
   * 
   * @param qname The qname.
   * @param qclass The qclass
   * @return A DS response message, if one could be constructed, null if not.
   */
  private SMessage lookupDSMessage(Name qname, int qclass)
  {
    RRsetEntry ds_entry = lookupRRset(rrsetKey(qname, Type.DS, qclass));
    if (ds_entry == null || !ds_entry.rrset.isSigned())
    {
      return null;
    }

    // Otherwise, we get to construct a DS response.
    SMessage m = new SMessage();

    // ANSWER section
    m.addRRset(ds_entry.rrset, Section.ANSWER);

    // AUTHORITY section
    // The NS set (if still in cache) will always be named the same as the
    // DS's signer (which is required to be the name of the zone that DS is
    // authoritative in).
    Name signerName = ds_entry.rrset.getSignerName();
    RRsetEntry ns_entry = lookupRRset(rrsetKey(signerName, Type.NS, qclass));
    if (ns_entry != null)
    {
      m.addRRset(ns_entry.rrset, Section.AUTHORITY);
    }

    // ADDITIONAL section
    // Just the A/AAAA records that go with the NS rrset (if we had it)
    if (ns_entry != null)
    {
      for (Iterator i = ns_entry.rrset.rrs(); i.hasNext();)
      {
        NSRecord ns = (NSRecord) i.next();
        RRsetEntry entry = lookupRRset(rrsetKey(ns.getTarget(),
            Type.A,
            qclass));
        if (entry != null)
        {
          m.addRRset(entry.rrset, Section.ADDITIONAL);
        }
        entry = lookupRRset(rrsetKey(ns.getTarget(), Type.AAAA, qclass));
        if (entry != null)
        {
          m.addRRset(entry.rrset, Section.ADDITIONAL);
        }
      }
    }

    return m;
  }

  private SMessage lookupDNSKEYMessage(Name qname, int qclass)
  {
    RRsetEntry dnskey_entry = lookupRRset(rrsetKey(qname, Type.DNSKEY, qclass));
    if (dnskey_entry == null || !dnskey_entry.rrset.isSigned())
    {
      return null;
    }

    log.debug("Synthesizing DNSKEY response -- dnskey rrset is signed: "
        + dnskey_entry.rrset.firstSig());
    // Otherwise, we get to construct a DS response.
    SMessage m = new SMessage();

    // ANSWER section
    m.addRRset(dnskey_entry.rrset, Section.ANSWER);

    // AUTHORITY section
    // The NS set (if still in cache) will have the same name as the DNSKEY.
    RRsetEntry ns_entry = lookupRRset(rrsetKey(qname, Type.NS, qclass));
    if (ns_entry != null)
    {
      m.addRRset(ns_entry.rrset, Section.AUTHORITY);
    }

    // ADDITIONAL section
    // Just the A/AAAA records that go with the NS rrset (if we had it)
    if (ns_entry != null)
    {
      for (Iterator i = ns_entry.rrset.rrs(); i.hasNext();)
      {
        NSRecord ns = (NSRecord) i.next();
        RRsetEntry entry = lookupRRset(rrsetKey(ns.getTarget(),
            Type.A,
            qclass));
        if (entry != null)
        {
          m.addRRset(entry.rrset, Section.ADDITIONAL);
        }
        entry = lookupRRset(rrsetKey(ns.getTarget(), Type.AAAA, qclass));
        if (entry != null)
        {
          m.addRRset(entry.rrset, Section.ADDITIONAL);
        }
      }
    }

    return m;
  }

  private long negativeCacheTTL(SMessage m, long minTTL)
  {
    SRRset[] rrsets = m.getSectionRRsets(Section.AUTHORITY, Type.SOA);
    if (rrsets == null || rrsets.length == 0)
    {
      if (mDefaultNegTTL < minTTL) return mDefaultNegTTL;
      return minTTL;
    }

    SOARecord soa = (SOARecord) rrsets[0].first();
    if (soa.getMinimum() < minTTL) return soa.getMinimum();
    return minTTL;
  }

  /**
   * Calculate the key for a cached RRset entry based on the DNS trinity.
   * 
   * @param n The name of the SRRset.
   * @param type The type of the SRRset.
   * @param dclass The class of the SRRset.
   * 
   * @return the key string.
   */
  private static String rrsetKey(Name n, int type, int dclass)
  {
    return "R" + type + "/" + dclass + "/" + Util.nameToString(n);
  }

  /**
   * Calculate the key for a cached RRset entry based on the RRset itself.
   * 
   * @param rrset The SRRset that will be cached.
   * 
   * @return the key string.
   */
  private static String rrsetKey(SRRset rrset)
  {
    return rrsetKey(rrset.getName(), rrset.getType(), rrset.getDClass());
  }

  /**
   * Lookup a given RRset in the cache using the given key.
   * 
   * @param k The key (created from rrsetKey() usually) to lookup.
   * 
   * @return a RRsetEntry (which will contain the cached SRRset) or null if 1)
   *         the key wasn't in the cache, or 2) the cache entry was expired.
   */
  private RRsetEntry lookupRRset(String k)
  {
    if (!k.startsWith("R"))
      throw new IllegalArgumentException("Wrong kind of key string used.");

    RRsetEntry entry = (RRsetEntry) mCache.get(k);
    if (entry == null) return null;

    // throw out the response if it has expired.
    if (entry.expiration != null && entry.expiration.before(new Date()))
    {
      log.debug("discarding expired rrset cache entry: " + entry
          + " (expired " + entry.expiration + ")");
      mCache.remove(k);
      return null;
    }

    // Clone the cached RRset, trimming the TTL in the process.
    SRRset orig_rrset = entry.rrset;
    long new_ttl = new_ttl = CacheUtils.cacheTTL(entry.expiration, orig_rrset
        .getTTL(), null);

    entry.rrset = orig_rrset.cloneSRRset(new_ttl);

    return entry;
  }

  /**
   * Given an SRRset to cache and a credibility rating, possibly store the
   * rrset in the cache.
   * 
   * @param rrset The SRRset to cache.
   * @param cred The credibility of this SRRset (basically, a number
   *          indicating where it came from).
   * 
   * @return the key used to store the rrset.
   */
  private String storeRRset(SRRset rrset, int cred)
  {
    String k = null;
    if (rrset == null) return k;
    if (rrset.getTTL() == 0) return k;

    // Trim the rrset TTL to match the RRSIG information, if necessary.
    CacheUtils.trimRRsetTTL(rrset, null);

    RRsetEntry entry = new RRsetEntry(rrset, mMaxTTL, cred);

    k = rrsetKey(rrset);

    RRsetEntry oldEntry = lookupRRset(k);

    // Store this new rrset if:
    // 1) there is no competing old entry.
    // 2) new rrset is signed and old entry isn't
    // 3) old rrset is secure and new rrset is identical (update TTL only,
    // really)
    // 4) new rrset has greater than or equal credibility, and has same level
    // of signedness.

    if (oldEntry == null)
    {
      mCache.put(k, entry);
      return k;
    }

    if (!oldEntry.rrset.isSigned() && rrset.isSigned())
    {
      log.debug("Replacing unsigned entry (" + oldEntry
          + ") with signed entry (" + entry + ")");
      mCache.put(k, entry);
      return k;
    }

    if (oldEntry.rrset.isSigned() && !rrset.isSigned())
    {
      log.debug("NOT replacing signed rrset with unsigned rrset: " + entry);
      return k;
    }

    if (oldEntry.rrset.getSecurityStatus() == SecurityStatus.SECURE)
    {
      if (rrset.equals(oldEntry.rrset)
          || rrset.getSecurityStatus() == SecurityStatus.SECURE)
      {
        log.debug("Replacing old secure rrset "
            + "with new equal or also secure rrset: " + entry);
        rrset.setSecurityStatus(SecurityStatus.SECURE);
        mCache.put(k, entry);
      }
      else
      {
        log.debug("NOT storing RRSET: old entry was SECURE "
            + "and new entry is different: " + oldEntry);
        log.debug("old rrset:\n" + oldEntry.rrset);
        log.debug("new different rrset:\n" + rrset);
      }
      return k;
    }

    if (oldEntry.credibility <= cred)
    {
      log.debug("Replacing old entry (" + oldEntry
          + ") with higher (or equal) credibility entry (" + entry + ")");
      mCache.put(k, entry);
      return k;
    }

    log.debug("NOT storing RRset: " + entry
        + " because credibility of old entry was " + oldEntry.credibility
        + " compared to " + entry.credibility);

    return k;
  }

  /**
   * Given a section number, return a rough estimate of the credibility of the
   * RRset in question.
   * 
   * @param section the section in question (ANSWER -> ADDITIONAL)
   * 
   * @return the credibility number.
   */
  private int credibility(int section, boolean auth_answer)
  {
    if (section == Section.ADDITIONAL) return Credibility.ADDITIONAL;
    if (section == Section.AUTHORITY)
    {
      if (auth_answer) return Credibility.AUTH_AUTHORITY;
      return Credibility.NONAUTH_AUTHORITY;
    }
    if (section == Section.ANSWER)
    {
      if (auth_answer) return Credibility.AUTH_ANSWER;
      return Credibility.NONAUTH_ANSWER;
    }
    // Otherwise
    return Credibility.HINT;
  }

  /**
   * Calculate the cache key for a message given the DNS trinity (this is
   * usually the same as the QUESTION section entry). This will generate the
   * key for a non-nxdomain cache entry.
   * 
   * @param n The qname
   * @param type The qtype
   * @param dclass The qclass
   * 
   * @return A key string.
   */
  private static String messageKey(Name n, int type, int dclass)
  {
    return "M" + type + "/" + dclass + "/" + Util.nameToString(n);
  }

  /**
   * Calculate the cache key for the given NXDOMAIN message.
   * 
   * @param n The name of the question.
   * @param dclass The class of the question.
   * 
   * @return A key string.
   */
  private static String nxMessageKey(Name n, int dclass)
  {
    return "N" + dclass + "/" + Util.nameToString(n);
  }

  /**
   * Given a key, lookup a cached message.
   * 
   * @param k The message key.
   * 
   * @return a MessageEntry, or null if the key wasn't present in the cache,
   *         or the message entry expired. Note that this isn't a "Message"
   *         itself, the contained SRRsets must still be fetched.
   */
  private MessageEntry lookupMessage(String k)
  {
    if (k.startsWith("R"))
      throw new IllegalArgumentException("Wrong kind of key string used.");

    MessageEntry entry = (MessageEntry) mCache.get(k);
    if (entry == null) return null;

    // throw out the response if it has expired.
    if (entry.expiration.before(new Date()))
    {
      log.debug("discarding expired message cache entry: " + entry
          + " (expired " + entry.expiration + ")");
      mCache.remove(k);
      return null;
    }

    return entry;
  }

  /**
   * Given a Message entry, fill out the entire message.
   * 
   * @param entry The message entry to start from.
   * 
   * @return A SMessage matching the message entry, or null if any of the
   *         message entry components were not in the cache.
   */
  private SMessage messageFromEntry(String messageKey, MessageEntry entry)
  {
    if (entry == null) return null;

    SMessage m = new SMessage(entry.header);
    m.setQuestion(entry.question);
    // this will allow the validator to set the security status of a message
    // entry directly.
    m.setSecurityStatus(entry.securityStatus);

    for (int s = Section.ANSWER; s <= Section.ADDITIONAL; s++)
    {
      String[] keys = entry.section[s - 1];
      if (keys == null) continue;
      for (int i = 0; i < keys.length; i++)
      {
        RRsetEntry re = lookupRRset(keys[i]);
        // If the RRset is not in the cache, we immediately expire the entire
        // message.
        if (re == null || re.rrset == null)
        {
          log.debug("discarding cached message because of missing rrset: "
              + keys[i]);
          mCache.remove(messageKey);
          return null;
        }
        // FIXME: we could add a credibility check here, but that shouldn't be
        // necessary (that is, we should never have too low of a credibility)
        m.addRRset(re.rrset, s);
      }
    }

    return m;
  }

}