/*
 * $Id: StandardHostCache.java 238 2005-08-15 22:26:43Z davidb $
 *
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.cache;

import java.net.InetAddress;
import java.util.Date;
import java.util.Properties;

import org.xbill.DNS.Name;

import se.rfc.unbound.Host;
import se.rfc.unbound.Util;

/**
 * This is a standard host cache implementation.
 * 
 * @author davidb
 * @version $Revision: 238 $
 */
public class StandardHostCache implements HostCache
{

  private static class Entry
  {
    public Date expiration;
    public Host host;

    public Entry(Host h, long ttl)
    {
      expiration = new Date(System.currentTimeMillis() + (ttl * 1000));
      host = h;
    }
  }

  // This is the main caching data structure.
  private CacheCore mCache;
  // This is the TTL that all host cache entries will have (in seconds)
  private long      mTTL          = 900;
  // This is the TTL that lameness entries will have (in seconds)
  private long      mLameTTL      = 900;
  // This is the maximum number of entries that the host cache will hold.
  private int       mMaxCacheSize = 1000;

  public StandardHostCache()
  {
    mCache = new CacheCore();
  }

  public void init(Properties config)
  {
    if (config == null) return;

    String s = config.getProperty("dns.hostcache.ttl");
    mTTL = Util.parseLong(s, mTTL);
    
    s = config.getProperty("dns.hostcache.max_size");
    mMaxCacheSize = Util.parseInt(s, mMaxCacheSize);
    mCache.setMaxCacheSize(mMaxCacheSize);
    
    s = config.getProperty("dns.hostcache.lame_ttl");
    mLameTTL = Util.parseLong(s, mLameTTL);
  }

  public Host lookup(InetAddress address)
  {
    Entry entry = (Entry) mCache.get(address);

    if (entry == null || entry.expiration.before(new Date()))
    {
      return store(new Host(address));
    }

    return entry.host;
  }

  private Host store(Host host)
  {
    Entry entry = new Entry(host, mTTL);
    // Replace any existing version of the Host object for now.
    mCache.put(host.getAddress(), entry);

    return entry.host;
  }

  private String lameKey(Host host, Name zone, int dclass)
  {
    return "L" + host.getAddress().getHostAddress() + "/"
        + Util.nameToString(zone) + "/" + dclass;
  }

  public void markLame(Host host, Name zone, int dclass)
  {
    Entry e = new Entry(host, mLameTTL);
    mCache.put(lameKey(host, zone, dclass), e);
  }

  public boolean isLame(Host host, Name zone, int dclass)
  {
    String k = lameKey(host, zone, dclass);
    Entry e = (Entry) mCache.get(k);
    
    if (e == null) return false;
    if (e.expiration.after(new Date()))
    {
      mCache.remove(k);
      return false;
    }
    
    return true;
  }
}
